from .parse_funcs import fbool, float01, float_or_str, int_or_path, lcstr, \
    floattuple_or_one, strlist_vsort, tupletupleint

config = {
    "bg": {
        "amplitude binary threshold":
            (None, float_or_str, "Binary image threshold value or method",
             "If not `None`, defines either a threshold for background "
             "segmentation or a method name in "
             ":const:`drymass.threshold.available_thresholds`."),
        "amplitude border perc":
            (10, float, "Amplitude bg border region to analyze [%]",
             "Set to 0 to disable."),
        "amplitude border px":
            (5, int, "Amplitude bg border region to analyze [px]",
             "Set to 0 to disable."),
        "amplitude data":
            (None, int_or_path, "Amplitude bg correction file or index",
             "Image indexing starts with 1. "
             "If a file, either specify the full path or the relative "
             "path to the directory containing the input data."),
        "amplitude mask sphere":
            (None, float, "Circular mask for amplitude bg correction",
             "If not `None`, a mask is used for background correction. "
             "A value of '1.0' results in a mask with the radius as "
             "determined with the edge-detection approach from phase [sic]. "
             "Set this to '1.1' to exclude peripheral phase values. "
             "If the amplitude border values or binary thresholds "
             "are set, the intersection of the resulting masks is used "
             "for background correction. For procedural details see "
             ":func:`qpsphere.cnvnc.bg_phase_mask_for_qpi`."),
        "amplitude offset":
            ("mean", lcstr, "Amplitude bg correction offset method",
             "Valid values are defined in "
             ":data:`qpimage.bg_estimate.VALID_FIT_OFFSETS`."),
        "amplitude profile":
            ("tilt", lcstr, "Amplitude bg correction profile method",
             "Valid values are defined in "
             ":data:`qpimage.bg_estimate.VALID_FIT_PROFILES`."),
        "enabled":
            (True, fbool, "Enable bg correction globally",
             "Set to *False* to disable background correction."),
        "phase binary threshold":
            (None, float_or_str, "Binary image threshold value or method",
             "If not `None`, defines either a threshold for background ",
             "segmentation or a method name in "
             ":const:`drymass.threshold.available_thresholds`."),
        "phase border perc":
            (10, float, "Phase bg border region to analyze [%]",
             "Set to 0 to disable."),
        "phase border px":
            (5, int, "Phase bg border region to analyze [px]",
             "Set to 0 to disable."),
        "phase data":
            (None, int_or_path, "Phase bg correction file or index",
             "Image indexing starts with 1. "
             "If a file, either specify the full path or the relative "
             "path to the directory containing the input data."),
        "phase mask sphere":
            (None, float, "Circular mask for phase bg correction",
             "If not `None`, a mask is used for background correction. "
             "A value of '1.0' results in a mask with the radius as "
             "determined with the edge-detection approach from phase. "
             "Set this to '1.1' to exclude peripheral phase values. "
             "If the phase border values or binary thresholds "
             "are set, the intersection of the resulting masks is used "
             "for background correction. For procedural details see "
             ":func:`qpsphere.cnvnc.bg_phase_mask_for_qpi`."),
        "phase offset":
            ("mean", lcstr, "Phase bg correction offset method",
             "Valid values are defined in "
             ":data:`qpimage.bg_estimate.VALID_FIT_OFFSETS`."),
        "phase profile":
            ("tilt", lcstr, "Phase bg correction profile method",
             "Valid values are defined in "
             ":data:`qpimage.bg_estimate.VALID_FIT_PROFILES`."),
    },
    "holo": {
        "filter name":  # filter_name
            ("disk", str, "Filter name for sideband isolation"),
        "filter size":  # filter_size
            (1/3, float, "Filter size (fraction of the sideband frequency)"),
        "sideband":  # sideband
            (1, floattuple_or_one, "Sideband ±1 or frequency coordinates"),
    },
    "meta": {
        "medium index":
            (None, float, "Refractive index of the surrounding medium"),
        "pixel size um":
            (None, float, "Detector pixel size [µm]"),
        "wavelength nm":
            (None, float, "Imaging wavelength [nm]"),
    },
    "output": {
        "roi images":
            (True, fbool, "Rendered phase images with ROI location"),
        "sphere images":
            (True, fbool, "Phase/Intensity images for sphere analysis"),
        "sensor tif data":
            (True, fbool, "Phase/Amplitude sensor tif data"),
    },
    "roi": {
        "dist border px":
            (10, int, "Minimum distance of objects to image border [px]"),
        "eccentricity max":
            (0.7, float, "Allowed maximal eccentricity of the specimen"),
        "enabled":
            (True, fbool, "Perform automated search for ROIs",
             "If set to `False`, the file 'roi_slices.txt' "
             "must contain ROIs."),
        "exclude overlap px":
            (30.0, float, "Allowed distance between two objects [px]"),
        "force":
            (None, tupletupleint, "Force ROI coordinates (x1,x2,y1,y2) [px]"),
        "ignore data":
            (None, strlist_vsort, "Exclude images and ROIs from the analysis",
             "To exclude individual ROIs found in a previous run, simply "
             "type the ROI index, i.e. 'exclude = 1.0, 2.2'. You may "
             "additionally exclude a full sensor image (e.g. image 3) with "
             "'exclude = 1.0, 2.2, 3'."),
        "pad border px":
            (40, int, "Padding of object regions [px]"),
        "size variation":
            (0.5, float01, "Allowed variation relative to specimen size"),
        "threshold":
            ("li", float_or_str, "Threshold for phase data segmentation",
             "The threshold is defined via a name or as a number in [rad]. "
             "Valid names are given in "
             ":const:`drymass.threshold.available_thresholds`."),
    },
    "specimen": {
        "size um":
            (10, float, "Approximate diameter of the specimen [µm]",
             "This is used as the initial value for the sphere analysis."),
    },
    "sphere": {
        "edge coarse":
            (0.4, float, "Coarse edge detection filter size"),
        "edge fine":
            (0.1, float, "Fine edge detection filter size"),
        "edge clip radius min":
            (0.9, float, "Interior edge point filtering radius"),
        "edge clip radius max":
            (1.1, float, "Exterior edge point filtering radius"),
        "edge iter":
            (20, int, "Maximum number iterations for coarse edge detection"),
        "image fit range position":  # crel
            (0.05, float, "Fit interpolation range for radius"),
        "image fit range radius":  # rrel
            (0.05, float, "Fit interpolation range for radius"),
        "image fit range refractive index":  # nrel
            (0.10, float, "Fit interpolation range for refractive index"),
        "image fix phase offset":  # fix_pha_offset
            (True, fbool, "Fix the simulation background phase to zero"),
        "image iter":  # max_iter
            (100, int, "Maximum number of iterations for image fitting"),
        "image stop delta position":  # stop_dc
            (1, float, "Stopping criterion for position"),
        "image stop delta radius":  # stop_dr
            (0.0010, float, "Stopping criterion for radius"),
        "image stop delta refractive index":  # stop_dn
            (0.0005, float, "Stopping criterion for refractive index"),
        "image verbosity":  # verbose
            (1, int, "Verbosity level of image fitting algorithm"),
        "method":
            ("edge", lcstr, "Method for determining sphere parameters",
             "Valid values are 'edge' (edge-detection approach) or "
             "'image' (2D phase image fitting)."),
        "model":
            ("projection", lcstr, "Physical sphere model",
             "Valid values are defined in "
             ":data:`qpsphere.models.available`. If `method=edge`, then "
             "`model` must be set to `projection`. If `method=image`, "
             "setting `model` to `rytov-sc` has the best trade-off between "
             "accuracy and speed."),
        "refraction increment":
            (0.18, float, "Refraction increment [mL/g]"),
        "radial inclusion factor":
            (1.2, float, "Radial inclusion factor for dry mass computation"),
    },
}



"""
There are two types of thresholding done in DryMass.

1. Thresholding of the `sensor image` which is required by the
   detection of the phase object ROIs
   (``[roi]: threshold`` configuration parameter)
2. Thresholding of the `individual ROIs` for determining the masked
   area used for ROI background correction
   (``[bg]: amplitude binary threshold``
   and ``[bg]: phase binary threshold`` configuration parameters)
"""
import warnings

import numpy as np
import skimage.filters as skfilters


def image2mask(image, value_or_method, invert=False):
    """Convert an image to a binary mask for background correction

    If `invert` is False, the threshold value is included in the
    resulting array.

    Parameters
    ----------
    image: 2d np.ndarray
        Input image
    value_or_method: float or str
        Either a threshold value or a string naming a
        filter method in :mod:`skimage.filters`.
    invert: bool
        Invert the resulting boolean array
    """
    if isinstance(value_or_method, str):
        if value_or_method.startswith("threshold_"):
            value_or_method = value_or_method[10:]
            warnings.warn("Use of the name of the thresholding function "
                          "from skimage is deprecated. Please use "
                          " '{}' instead!".format(value_or_method),
                          DeprecationWarning)
        method = threshold_dict[value_or_method]
        bw = image >= method(image)
    else:
        bw = image >= value_or_method
    if invert:
        return ~bw
    else:
        return bw


def threshold_drymass_nuclei(image):
    """Threshold filter for segmenting cell nuclei in phase images

    Cell nuclei have a low refractive index, but the nucleoli within
    the nuclei usually have a very high refractive index. As a result,
    conventional thresholding algorithms either cannot detect the
    nuclei reliably or segment the nucleoli only.

    This thresholding filter copes with the situation by
    "pulling down" the top 1% of the phase data and taking the
    threshold at 20% of the maximum phase relative to the mean
    of the original phase data.
    """
    image = image.copy()
    mean = np.mean(image)
    size = image.size
    # ignore the top 1%
    counter = 0
    maxim = np.max(image)
    while counter < size // 100:
        maxid = image == maxim
        counter += np.sum(maxid)
        maxim = np.max(image[~maxid])
        image[maxid] = maxim
    # take 20% as threshold value
    thresh = mean + .2 * (maxim-mean)
    return thresh


def threshold_li(image):
    """Li threshold optimized for cells in QPI"""
    return skfilters.threshold_li(image,
                                  initial_guess=np.percentile(image, q=95))


#: Dictionary containing all thresholding methods available in DryMass
threshold_dict = {
    'dm-nuclei': threshold_drymass_nuclei,
    'isodata': skfilters.threshold_isodata,
    'li': threshold_li,
    'mean': skfilters.threshold_mean,
    'minimum': skfilters.threshold_minimum,
    'otsu': skfilters.threshold_otsu,
    'triangle': skfilters.threshold_triangle,
    'yen': skfilters.threshold_yen,
    }

#: Available thresholding method names;
#: The thresholding methods are either defined in this module
#: (see `threshold_*` methods) or taken from :mod:`skimage.filters`.
available_thresholds = sorted(threshold_dict.keys())



from .analyzing import cli_analyze_sphere  # noqa: F401
from .converting import cli_convert  # noqa: F401
from .extracting import cli_extract_roi  # noqa: F401
from .profile import cli_profile  # noqa: F401



"""Parse DryMass definitions.py file"""
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes

from drymass.cli.definitions import config


class IncludeDirective(Directive):
    required_arguments = 1
    optional_arguments = 0

    def run(self):
        sec = self.arguments[0]

        content = config[sec]

        rst = []

        for key in content:
            # get data
            if len(content[key]) == 3:
                default, func, doc = content[key]
                notes = None
            elif len(content[key]) == 4:
                default, func, doc, notes = content[key]
            # parse dtype
            dtype = func.__name__
            if dtype in ["float", "int", "str"]:
                str_dtype = ":class:`{}`".format(dtype)
            else:
                str_dtype = ":func:`{0} <drymass.cli.parse_funcs.{0}>`".format(
                            dtype)
            # add list item
            rst.append("* | **{}** = {} ({}) -- {} ".format(key, default,
                                                            str_dtype, doc))
            if notes is not None:
                # add notes
                rst.append("  | {}".format(notes.strip()))
            rst.append("")

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


def setup(app):
    app.add_directive('include_definition', IncludeDirective)

    return {'version': '0.1'}   # identifies the version of our extension



import argparse
import pathlib
import shutil

import appdirs


APP_DIR = pathlib.Path(appdirs.user_config_dir(appname="drymass"))


def cli_profile(args=None):
    if args is None:
        parser = profile_parser()
        args = parser.parse_args()
    cmd = args.subparser_name
    APP_DIR.mkdir(parents=True, exist_ok=True)
    if cmd in [None, "list"]:
        # default to listing all profiles
        profiles = get_profiles()
        if profiles:
            print("Available profiles:")
            for p in profiles:
                name = p.name.split("_", 1)[1][:-4]
                print(" - {}: {}".format(name, p))
        else:
            print("No profiles in local library.")
    elif cmd == "add":
        loc_path = APP_DIR / "profile_{}.cfg".format(args.name)
        if loc_path.exists():
            raise OSError("Profile '{}' already exists!".format(args.name))
        shutil.copy(args.path, loc_path)
    elif cmd == "remove":
        loc_path = APP_DIR / "profile_{}.cfg".format(args.name)
        if loc_path.exists():
            loc_path.unlink()
        else:
            raise OSError("Profile '{}' does not exist!".format(args.name))
    elif cmd == "export":
        paths = get_profiles()
        out_path = pathlib.Path(args.path)
        out_path.mkdir(parents=True, exist_ok=True)
        for p in paths:
            shutil.copy(p, out_path)
    else:
        raise ValueError("Invalid command '{}'!".format(cmd))


def get_profiles():
    """Return the paths to all profiles in the local library"""
    paths = APP_DIR.glob("profile_*")
    return sorted(paths)


def get_profile_path(name):
    """Convenience function for obtaining a profile path

    If `name` is a path to a profile file, then this path is
    returned. If `name` is a name in the local profile library,
    then the path to that profile is returned.
    """
    path = pathlib.Path(name)
    if not path.exists():
        # name from local library
        path = APP_DIR / "profile_{}.cfg".format(name)
    return path


def profile_parser():
    parser = argparse.ArgumentParser(
        description="Management of DryMass profiles. "
                    "Profiles are just drymass.cfg files that contain "
                    "user-defined default values. Profiles are stored "
                    "in a local library located in the user's configuration "
                    "directory.")
    subparsers = parser.add_subparsers(
        title="subcommands",
        description="Run `dm_profile subcommand --help` for "
        "more information.",
        dest='subparser_name')

    t_add = "adds a new profile to the local library"
    p_add = subparsers.add_parser("add", help=t_add, description=t_add)
    p_add.add_argument("name", type=str, help="profile name")
    p_add.add_argument("path", type=str, help="path to drymass.cfg")

    t_rem = "removes a profile from the local library"
    p_rem = subparsers.add_parser("remove", help=t_rem, description=t_rem)
    p_rem.add_argument("name", type=str, help="profile name")

    t_list = "list all profiles in the local library"
    subparsers.add_parser("list", help=t_list, description=t_list)

    t_exp = "export all profile in the local library"
    p_exp = subparsers.add_parser("export", help=t_exp, description=t_exp)
    p_exp.add_argument("path", type=str, help="export path")

    return parser
